import asyncio
import json

import asyncio_mqtt as aiomqtt

from .. import fhem, generic, utils


class mqtt_ha_discovery(generic.FhemModule):

    existing_devices = []

    def __init__(self, logger):
        super().__init__(logger)

        self.reset_existing_devices_task = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {
            "force_autocreate": {
                "default": False,
                "format": "bool",
                "help": "Always create a new device on discovery.",
            }
        }
        await self.set_attr_config(attr_config)

        set_config = {}
        await self.set_set_config(set_config)

        if len(args) > 7:
            return "Usage: define mqttdiscovery fhempy mqtt_ha_discovery"

        if len(args) > 3:
            self.mqtt_host = args[3]
        else:
            self.mqtt_host = "127.0.0.1"

        if len(args) > 4:
            self.mqtt_port = int(args[4])
        else:
            self.mqtt_port = 1883

        if len(args) > 5:
            self.mqtt_user = args[5]
        else:
            self.mqtt_user = None

        if len(args) > 6:
            self.mqtt_pw = args[6]
        else:
            self.mqtt_pw = None

        self.create_async_task(self.setup())

    async def setup(self):
        await self.connect_to_broker()

    async def connect_to_broker(self):
        try:
            async with aiomqtt.Client(
                self.mqtt_host,
                self.mqtt_port,
                username=self.mqtt_user,
                password=self.mqtt_pw,
                logger=self.logger,
            ) as client:
                async with client.messages() as messages:
                    await client.subscribe("homeassistant/#")
                    async for message in messages:
                        await self.handle_ha_msg(message)
        except aiomqtt.MqttError:
            self.logger.error("Connection lost; Reconnecting in 5 seconds ...")
            await asyncio.sleep(5)

    async def handle_ha_msg(self, msg):
        topic = msg.topic.value
        payload = msg.payload
        payload_json = json.loads(payload)

        device_name = payload_json["dev"]["name"]
        fhem_device_name = await self.create_device(topic, device_name, payload_json)
        await self.set_attributes(fhem_device_name, topic, payload_json)

    async def set_attributes(self, fhem_device_name, topic, payload_json):
        cur_readingList = await fhem.AttrVal(fhem_device_name, "readingList", "")
        cur_setList = await fhem.AttrVal(fhem_device_name, "setList", "")

        # split topic
        _, component, dev_name, object_id, topic_type = topic.split("/")

        if topic_type != "config":
            return

        # get attribute based on topic
        readingList = None
        setList = None
        if component == "switch":
            readingList, setList = await self.generate_switch_attributes(
                object_id, topic, payload_json
            )
        elif component == "light":
            readingList, setList = await self.generate_light_attributes(
                object_id, topic, payload_json
            )
        elif component == "sensor":
            readingList, setList = await self.generate_sensor_attributes(
                object_id, topic, payload_json
            )

        # check if topic is already in the attribute
        if readingList and cur_readingList.find(payload_json.get("stat_t")) == -1:
            if cur_readingList != "":
                cur_readingList += "\n"
            new_readingList = cur_readingList + readingList
            await fhem.CommandAttr(
                self.hash, f"{fhem_device_name} readingList {new_readingList}"
            )

        if setList and cur_setList.find(payload_json.get("cmd_t")) == -1:
            if cur_setList != "":
                cur_setList += "\n"
            new_setList = cur_setList + setList
            await fhem.CommandAttr(
                self.hash, f"{fhem_device_name} setList {new_setList}"
            )

    async def generate_switch_attributes(self, object_id, topic, payload_json):
        # Example
        # {
        #     "name": "yellow",
        #     "stat_t": "baerli/switch/yellow/state",
        #     "cmd_t": "baerli/switch/yellow/command",
        #     "avty_t": "baerli/status",
        #     "uniq_id": "ESPswitchyellow",
        #     "dev": {
        #         "ids": "b4e62d451d0d",
        #         "name": "baerli",
        #         "sw": "esphome v2022.8.3 Oct  8 2022, 08:35:20",
        #         "mdl": "d1_mini",
        #         "mf": "espressif",
        #     },
        # }

        # create readingList entry
        reading_list = [
            payload_json["stat_t"]
            + ":.* { { "
            + payload_json["name"]
            + "=>lc(\$EVENT) } }"
        ]
        # create setList entry
        set_list = [
            payload_json["name"] + "_on " + payload_json["cmd_t"] + " on",
            payload_json["name"] + "_off " + payload_json["cmd_t"] + " off",
        ]
        return ("\n".join(reading_list), "\n".join(set_list))

    async def generate_light_attributes(self, object_id, topic, payload_json):
        # Example
        # {
        #     "schema": "json",
        #     "clrm": True,
        #     "supported_color_modes": ["brightness"],
        #     "brightness": True,
        #     "name": "Light",
        #     "stat_t": "baerli/light/light/state",
        #     "cmd_t": "baerli/light/light/command",
        #     "avty_t": "baerli/status",
        #     "uniq_id": "ESPlightlight",
        #     "dev": {
        #         "ids": "b4e62d451d0d",
        #         "name": "baerli",
        #         "sw": "esphome v2022.8.3 Oct  8 2022, 08:35:20",
        #         "mdl": "d1_mini",
        #         "mf": "espressif",
        #     },
        # }

        # create readingList entry
        reading_list = [
            payload_json["stat_t"]
            + ":.* { { "
            + payload_json["name"]
            + "=>lc(\$EVENT) } }"
        ]
        # create setList entry
        set_list = [
            payload_json["name"]
            + "_on "
            + payload_json["cmd_t"]
            + '{"state": "on", "transition": 1}',
            payload_json["name"]
            + "_off "
            + payload_json["cmd_t"]
            + '{"state": "on", "transition": 1}',
        ]
        if "brightness" in payload_json["supported_color_modes"]:
            set_list.append(
                payload_json["name"]
                + "_brightness:colorpicker,BRI,0,5,255 "
                + payload_json["cmd_t"]
                + '{"state":"on","\$EVTPART0":"\$EVTPART1"}'
            )
        return ("\n".join(reading_list), "\n".join(set_list))

    async def generate_sensor_attributes(self, object_id, topic, payload_json):
        # Example
        # {
        #     "dev_cla": "voltage",
        #     "unit_of_meas": "V",
        #     "stat_cla": "measurement",
        #     "name": "luminance",
        #     "stat_t": "airpurifier/sensor/luminance/state",
        #     "avty_t": "airpurifier/status",
        #     "uniq_id": "807d3a78f153-adc",
        #     "dev": {
        #         "ids": "807d3a78f153",
        #         "name": "airpurifier",
        #         "sw": "esphome v2022.9.4 Nov 19 2022, 09:04:05",
        #         "mdl": "d1_mini",
        #         "mf": "espressif",
        #     },
        # }

        # create readingList entry
        reading_list = [
            payload_json["stat_t"]
            + ":.* { { "
            + payload_json["name"]
            + "=>lc(\$EVENT) } }"
        ]
        return ("\n".join(reading_list), None)

    async def reset_existing_devices(self):
        await asyncio.sleep(60)
        mqtt_ha_discovery.existing_devices.clear()

    async def create_device(self, topic, device_name, payload_json):
        if device_name in mqtt_ha_discovery.existing_devices:
            return

        dev_exists = await fhem.CommandList(
            self.hash, "readingList=.*" + device_name + "/.* NAME"
        )
        if dev_exists:
            mqtt_ha_discovery.existing_devices.append(device_name)
            if self.reset_existing_devices_task is None:
                self.reset_existing_devices_task = self.create_async_task(
                    self.reset_existing_devices()
                )
            return dev_exists.split()[0]

        fhem_name = utils.gen_fhemdev_name(
            "MQTT2_" + device_name + "_" + payload_json["dev"]["ids"]
        )

        await fhem.CommandDefine(self.hash, f"{fhem_name} MQTT2_DEVICE")
        return fhem_name



import asyncio
import binascii
import concurrent.futures
import inspect
import json
import os
import socket
from base64 import urlsafe_b64decode, urlsafe_b64encode
from codecs import decode
from datetime import datetime
from functools import partial, reduce

from Cryptodome.Cipher import AES
from Cryptodome.Hash import HMAC, SHA256
from Cryptodome.Util.Padding import unpad

from . import fhem


def encrypt_string(plain_text, fhem_unique_id):
    key = fhem_unique_id.encode("utf-8")
    b_text = plain_text.encode("utf-8")
    e_cipher = AES.new(key, AES.MODE_EAX, nonce=key[0:16])
    e_data = e_cipher.encrypt(b_text)
    return "crypt-aes:" + urlsafe_b64encode(e_data).decode()


def decrypt_string(encrypted_text, fhem_unique_id):
    if encrypted_text[0:10] != "crypt-aes:":
        return decrypt_fernet(encrypted_text, fhem_unique_id)
    key = fhem_unique_id.encode("utf-8")
    encrypted_data = urlsafe_b64decode(encrypted_text[10:])
    d_cipher = AES.new(key, AES.MODE_EAX, nonce=key[0:16])
    return d_cipher.decrypt(encrypted_data).decode()


def decrypt_fernet(token_b64, key_str):
    try:
        keys = key_str.encode("utf-8")
        token_z = token_b64.encode("utf-8")
        token_b64 = reduce(decode, ("base64", "zlib"), token_z).decode()
        token = urlsafe_b64decode(token_b64)

        if len(keys) != 32 or len(token) < 73 or token[0] != 0x80:
            raise ValueError("Incorrect format")

        hmac_key, dec_key = keys[:16], keys[16:]
        HMAC.new(hmac_key, token[:-32], digestmod=SHA256).verify(token[-32:])
        iv, ct = token[9:25], token[25:-32]
        # Possibly check time-to-live here
        message = unpad(AES.new(dec_key, AES.MODE_CBC, iv=iv).decrypt(ct), 16).decode()

    except (ValueError, binascii.Error):
        message = None

    return message


def get_fhempy_root():
    from fhempy import lib

    initfile = inspect.getfile(lib)
    fhempy_root = os.path.dirname(initfile)
    return fhempy_root


async def run_blocking(function):
    if isinstance(function, partial) is False:
        raise Exception("Use functools.partial to call run_blocking")

    with concurrent.futures.ThreadPoolExecutor() as pool:
        return await asyncio.get_event_loop().run_in_executor(pool, function)


def run_blocking_task(function):
    return asyncio.create_task(run_blocking(function))


# example config
# attr_list = {
#   "attribute1": {"default": 10, "format": "int", "options":"1,2,3"}
# }
async def handle_attr(attr_list, obj, hash, args, argsh):
    cmd = args[0]
    name = args[1]
    attr_name = args[2]
    attr_val = args[3]
    if attr_name in attr_list:
        if cmd == "set":
            setattr(
                obj,
                "_attr_" + attr_name,
                convert2format(attr_val, attr_list[attr_name]),
            )
        else:
            if "default" in attr_list[attr_name]:
                setattr(obj, "_attr_" + attr_name, attr_list[attr_name]["default"])
            else:
                setattr(obj, "_attr_" + attr_name, "")

    # call set_attr_....
    fct_name = "set_attr_" + attr_name
    ret = None
    if attr_name in attr_list and "function" in attr_list[attr_name]:
        fct_name = attr_list[attr_name]["function"]
    try:
        fct_call = getattr(obj, fct_name)
        ret = await fct_call(hash)
    except AttributeError:
        pass

    return ret


def get_local_ip():
    sock = None
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        # Use Google Public DNS server to determine own IP
        sock.connect(("8.8.8.8", 80))

        return sock.getsockname()[0]  # type: ignore
    except OSError:
        try:
            return socket.gethostbyname(socket.gethostname())
        except socket.gaierror:
            return "127.0.0.1"
    finally:
        if sock is not None:
            sock.close()


async def handle_define_attr(attr_list, obj, hash):
    add_to_list = []
    for attr in attr_list:
        if "options" in attr_list[attr]:
            attr_opt = attr + ":" + attr_list[attr]["options"]
        else:
            attr_opt = attr
        add_to_list.append(attr_opt)
    await fhem.setDevAttrList(hash["NAME"], " ".join(add_to_list))

    for attr in attr_list:
        curr_val = await fhem.AttrVal(hash["NAME"], attr, "")
        if curr_val == "" and "default" in attr_list[attr]:
            curr_val = attr_list[attr]["default"]
        setattr(obj, "_attr_" + attr, convert2format(curr_val, attr_list[attr]))

        # call set_attr_....
        # fct_name = "set_attr_" + attr
        # if "function" in attr_list[attr]:
        #    fct_name = attr_list[attr]["function"]
        # try:
        #    fct_call = getattr(obj, fct_name)
        #    await fct_call(hash)
        # except AttributeError:
        #    pass

    return


def flatten_json(y):
    out = {}
    if type(y) is str:
        y = json.loads(y)

    def flatten(x, name=""):
        if type(x) is dict:
            for a in x:
                flatten(x[a], name + a + "_")
        elif type(x) is list:
            i = 0
            for a in x:
                flatten(a, name + str(i) + "_")
                i += 1
        else:
            out[name[:-1]] = x

    flatten(y)
    return out


def gen_reading_name(reading):
    reading = reading.lower()
    reading = reading.replace(" ", "_").replace("-", "_")
    return remove_special_charachters(reading)


def gen_fhemdev_name(devname):
    return remove_special_charachters(
        devname.replace(" ", "_").replace("-", "_").replace(":", "_")
    )


def remove_special_charachters(string):
    """
    Removes umlauts from strings and replaces them with the letter+e convention
    :param string: string to remove umlauts from
    :return: unumlauted string
    """
    u_enc = "ü".encode()
    U_enc = "Ü".encode()
    a_enc = "ä".encode()
    A_enc = "Ä".encode()
    o_enc = "ö".encode()
    O_enc = "Ö".encode()
    ss_enc = "ß".encode()

    string = string.encode()
    string = string.replace(u_enc, b"ue")
    string = string.replace(U_enc, b"Ue")
    string = string.replace(a_enc, b"ae")
    string = string.replace(A_enc, b"Ae")
    string = string.replace(o_enc, b"oe")
    string = string.replace(O_enc, b"Oe")
    string = string.replace(ss_enc, b"ss")

    string = string.decode("utf-8")

    string = string.replace("(", "")
    string = string.replace(")", "")
    string = string.replace("'", "")
    string = string.replace("`", "")
    string = string.replace('"', "")
    return string


def convert2format(value, list_def):
    target_format = "none"
    if "format" in list_def:
        target_format = list_def["format"]

    if target_format == "int":
        return int(value)
    elif target_format == "float":
        return float(value)
    elif target_format == "str":
        return str(value)
    elif target_format == "bool":
        if value == "on":
            return True
        else:
            return False
    elif target_format == "json":
        try:
            return json.loads(value)
        except Exception:
            return {}
    elif target_format == "time":
        try:
            return datetime.strptime(value, "%H:%M")
        except Exception:
            return None
    elif target_format == "array":
        return value.split(",")
    return value


# example config
# set_list_conf = {
#    "mode": { "args": ["mode"], "argsh": ["mode"], "params": { "mode": { "default": "eco", "optional": False }}, "options": "eco,comfort" },
#    "desiredTemp": { "args": ["temperature"], "options": "slider,10,1,30"},
#    "holidayMode": { "args": ["start", "end", "temperature"], "params": { "start": {"default": "Monday"}, "end": {"default": "23:59"}}},
#    "on": { "args": ["seconds"], "params": { "seconds": {"optional": True}}},
#    "off": {}
# }
async def handle_set(set_list_conf, obj, hash, args, argsh):
    fhem_set_list = []
    if len(args) < 2 or (len(argsh) == 0 and args[1] == "?"):
        for cmd in set_list_conf:
            if "options" in set_list_conf[cmd]:
                fhem_options = ":" + set_list_conf[cmd]["options"]
            elif "args" in set_list_conf[cmd] or "argsh" in set_list_conf[cmd]:
                fhem_options = ""
            else:
                fhem_options = ":noArg"
            fhem_set_list.append(cmd + fhem_options)
        return "Unknown argument ?, choose one of " + " ".join(fhem_set_list)
    else:
        # get cmd
        cmd = args[1]
        if cmd in set_list_conf:
            all_args = {}
            if "argsh" in set_list_conf[cmd]:
                all_args = argsh
            cmd_def = set_list_conf[cmd]
            # map arguments to params
            # add args to all_args
            if ("args" in cmd_def and (len(args) - 2) > len(cmd_def["args"])) or (
                len(args) > 2 and args[2] == "?"
            ):
                return f"Usage: set {hash['NAME']} {cmd} " + " ".join(cmd_def["args"])
            i = 0
            for arg in args[2:]:
                # arg ... mode
                # all_args[mode] = mode argument
                if "args" in cmd_def and i < len(cmd_def["args"]):
                    all_args[cmd_def["args"][i]] = arg
                else:
                    return f"Too many parameters provided: {arg}"
                i += 1
            # get default values for other params
            final_params = all_args
            if "params" in cmd_def:
                # add value to params
                for arg in all_args:
                    if arg in cmd_def["params"]:
                        cmd_def["params"][arg]["value"] = all_args[arg]
                for param in cmd_def["params"]:
                    # check if value is available or default value
                    # check if all required params are availble
                    if "value" in cmd_def["params"][param]:
                        final_params[param] = cmd_def["params"][param]["value"]
                    elif "default" not in cmd_def["params"][param] and (
                        "optional" not in cmd_def["params"][param]
                        or cmd_def["params"][param]["optional"] is False
                    ):
                        # no value found, check if optional
                        return f"Required argument {param} missing."
                    elif (
                        "default" in cmd_def["params"][param]
                        and "value" not in cmd_def["params"][param]
                    ):
                        final_params[param] = cmd_def["params"][param]["default"]
                    if "format" in cmd_def["params"][param] and param in final_params:
                        final_params[param] = convert2format(
                            final_params[param], cmd_def["params"][param]
                        )

            # call function with params
            if "function_param" in set_list_conf[cmd]:
                final_params["function_param"] = set_list_conf[cmd]["function_param"]
            if "function" in set_list_conf[cmd]:
                fct_name = set_list_conf[cmd]["function"]
                final_params["cmd"] = cmd
            else:
                fct_name = "set_" + cmd
            fct_call = getattr(obj, fct_name)
            return await fct_call(hash, final_params)
        else:
            return "Command not available for this device."



#!/usr/bin/env python3
"""Support for Tuya Smart devices."""

import fhempy.lib.tuya_smartlife.tuya_smartlife_device as tsd
import fhempy.lib.tuya_smartlife.tuya_smartlife_setup as tss
from fhempy.lib import fhem, utils
from fhempy.lib.generic import FhemModule


class tuya_smartlife(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.device = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        # initialize set commands
        set_conf = {
            "scan_done": {},
            "relogin": {},
        }
        await self.set_set_config(set_conf)

        # define my_tuya_smartlife fhempy tuya_smartlife USERCODE
        # check if len(args) is between 4 and 5
        if not (4 <= len(args) <= 5):
            return (
                "Usage: define my_tuya_smartlife fhempy tuya_smartlife setup USERCODE"
            )

        if args[3] == "setup":
            self.usercode = args[4]

            self.setup_device = tss.tuya_smartlife_setup(self.logger, self)

            # start login procedure asynchron
            self.create_async_task(self.setup_device.login())
        else:
            self.device = tsd.tuya_smartlife_device(self.logger, self)
            await self.device.Define(self.hash, args, argsh)

    async def Undefine(self, hash):
        if self.device:
            await self.device.Undefine(hash)
        return await super().Undefine(hash)

    async def set_scan_done(self, hash, params):
        self.create_async_task(self.setup_device.set_scan_done(hash, params))

    async def set_relogin(self, hash, params):
        # clear token_info, terminal_id and endpoint
        await fhem.readingsSingleUpdate(self.hash, "token_info", "{}", 1)
        await fhem.readingsSingleUpdate(self.hash, "terminal_id", "", 1)
        await fhem.readingsSingleUpdate(self.hash, "endpoint", "", 1)
        self.create_async_task(self.setup_device.login())

    async def set_reset_energy(self, hash, params):
        await self.device.set_reset_energy(hash, params)

    async def set_boolean(self, hash, params):
        await self.device.set_boolean(hash, params)

    async def set_enum(self, hash, params):
        await self.device.set_enum(hash, params)

    async def set_json(self, hash, params):
        await self.device.set_json(hash, params)

    async def set_string(self, hash, params):
        await self.device.set_string(hash, params)

    async def set_integer(self, hash, params):
        await self.device.set_integer(hash, params)

    async def set_colour_data(self, hash, params):
        await self.device.set_colour_data(hash, params)

    async def set_colour_data_v2(self, hash, params):
        await self.device.set_colour_data_v2(hash, params)

    @property
    def tuya_cloud_device(self):
        return self.device



import asyncio

from warema_wms import Shade, WmsController

from .. import fhem, utils
from ..generic import FhemModule


class warema(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.hash = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        attr_config = {
            "interval": {
                "default": 60,
                "format": "int",
                "help": "Change interval, default is 60.",
            },
        }
        await self.set_attr_config(attr_config)

        set_config = {
            "status": {},
            "up": {},
            "down": {},
            "position": {"args": ["position"], "options": "slider,0,11,100"},
        }
        await self.set_set_config(set_config)

        self.hash = hash
        if len(args) < 5:
            return "Usage: define warema_fhempy fhempy warema <IP> <channel>"

        if await fhem.AttrVal(self.hash["NAME"], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon fts_window_1w")
        if await fhem.AttrVal(self.hash["NAME"], "devStateIcon", "") == "":
            devStateIcon = "0:fts_shutter_10\@green 100:fts_shutter_100\@green s/[0-9]|1\d.*:fts_shutter_90 s/[0-9]|2\d.*:fts_shutter_80 s/[0-9]|3\d.*:fts_shutter_70 s/[0-9]|4\d.*:fts_shutter_60 s/[0-9]|5\d.*:fts_shutter_50 s/[0-9]|6\d.*:fts_shutter_40 s/[0-9]|7\d.*:fts_shutter_30 s/[0-9]|8\d.*:fts_shutter_20 s/[0-9]|9\d.*:fts_shutter_10 s/[0-9]|99\d.*:fts_shutter_10"
            await fhem.CommandAttr(
                self.hash, self.hash["NAME"] + " devStateIcon " + devStateIcon
            )
        if await fhem.AttrVal(self.hash["NAME"], "webCmd", "") == "":
            await fhem.CommandAttr(
                self.hash, self.hash["NAME"] + " webCmd up:down:status"
            )
        if await fhem.AttrVal(self.hash["NAME"], "stateFormat", "") == "":
            await fhem.CommandAttr(
                self.hash, self.hash["NAME"] + " stateFormat position"
            )
        # if await fhem.AttrVal(self.hash["NAME"], "verbose", "") == "":
        #     await fhem.CommandAttr(self.hash, self.hash["NAME"] + " verbose 5")

        self._warema_ip = args[3]
        hash["IP"] = args[3]

        self._warema_channel = int(args[4])
        self.hash["CHANNEL"] = args[4]

        self._warema_shades = Shade.get_all_shades(
            WmsController("http://" + self._warema_ip)
        )

        self._warema_room = self._warema_shades[self._warema_channel].get_room_name()
        self.hash["ROOM"] = self._warema_room

        state = self._warema_shades[self._warema_channel].get_shade_state()
        (position, ismoving, date) = state

        self._warema_position = str(int(position))
        self.hash["POSITION"] = self._warema_position

        self._warema_ismoving = ismoving
        self.hash["ISMOVING"] = self._warema_ismoving

        if self._warema_position == 0:
            pos = "open"
        elif self._warema_position == 100:
            pos = "closed"
        else:
            pos = self._warema_position

        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdate(hash, "state", pos)
        await fhem.readingsBulkUpdate(hash, "room", self._warema_room)
        await fhem.readingsBulkUpdate(hash, "channel", self._warema_channel)
        await fhem.readingsBulkUpdate(hash, "position", self._warema_position)
        await fhem.readingsBulkUpdate(hash, "ismoving", self._warema_ismoving)
        await fhem.readingsEndUpdate(hash, 1)

        self.updateTask = self.create_async_task(self.update_task())

    async def update_task(self):
        while True:
            await self.do_update()
            await asyncio.sleep(self._attr_interval)

    async def do_update(self):
        state = self._warema_shades[self._warema_channel].get_shade_state()
        (position, ismoving, date) = state

        self._warema_position = int(position)
        self._warema_ismoving = ismoving

        await self.updateDeviceReadings()
        return

    async def set_attr_interval(self, hash):
        await fhem.readingsSingleUpdate(
            self.hash, "interval", str(self._attr_interval), 1
        )
        self.updateTask = self.create_async_task(self.update_task())

    # Set functions in format: set_NAMEOFSETFUNCTION(self, hash, params)
    async def set_status(self, hash, params):
        self.create_async_task(self.do_update())

    async def set_up(self, hash, params):
        self._warema_shades[self._warema_channel].set_shade_position(
            0
        )  # 0=open; 100=closed
        state = self._warema_shades[self._warema_channel].get_shade_state(
            True
        )  # Force update and get shade state
        (position, ismoving, date) = state

        self._warema_position = str(int(position))
        self._warema_ismoving = ismoving

        await self.updateDeviceReadings()

    async def set_down(self, hash, params):
        self._warema_shades[self._warema_channel].set_shade_position(
            100
        )  # 0=open; 100=closed
        state = self._warema_shades[self._warema_channel].get_shade_state(
            True
        )  # Force update and get shade state
        (position, ismoving, date) = state

        self._warema_position = str(int(position))
        self._warema_ismoving = ismoving

        await self.updateDeviceReadings()

    async def set_position(self, hash, params):
        pos = int(params["position"])

        self._warema_shades[self._warema_channel].set_shade_position(
            pos
        )  # 0=open; 100=closed
        state = self._warema_shades[self._warema_channel].get_shade_state(
            True
        )  # Force update and get shade state
        (position, ismoving, date) = state

        self._warema_position = int(position)
        self._warema_ismoving = ismoving

        await self.updateDeviceReadings()

    async def updateDeviceReadings(self):
        self.hash["POSITION"] = self._warema_position
        self.hash["ISMOVING"] = self._warema_ismoving

        if self._warema_position == 0:
            pos = "open"
        elif self._warema_position == 100:
            pos = "closed"
        else:
            pos = self._warema_position

        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdate(self.hash, "state", pos)
        await fhem.readingsBulkUpdate(self.hash, "position", self._warema_position)
        await fhem.readingsBulkUpdate(self.hash, "ismoving", self._warema_ismoving)
        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio
import base64
import pathlib
from datetime import datetime

import aiohttp

from .. import fhem, generic


class github_restore(generic.FhemModule):
    """Restore FHEM backup from github repository"""

    headers = {
        "Authorization": "",
        "Accept": "application/vnd.github+json",
    }

    def __init__(self, logger):
        super().__init__(logger)
        self.gh_token_ready = asyncio.Event()

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {
            "github_token": {"default": "", "help": "Personal github token"},
        }
        await self.set_attr_config(attr_config)

        set_config = {
            "restore_now": {},
        }
        await self.set_set_config(set_config)

        if len(args) != 5:
            return (
                "Usage: define my_restore fhempy github_restore "
                + " https://github.com/xxx/fhem_backup master_fhem_rpi"
            )

        self.url = args[3]
        self.gh_user = self.url.split("/")[-2]
        self.gh_repo = self.url.split("/")[-1]
        self.directory = args[4]

        icon = await fhem.AttrVal(self.hash["NAME"], "icon", "")
        if icon == "":
            await fhem.CommandAttr(self.hash, f"{self.hash['NAME']} icon system_backup")
            await fhem.CommandAttr(
                self.hash,
                f"{self.hash['NAME']} devStateIcon "
                + "failed:message_attention ok:message_ok",
            )

        # do this to check if gh token is set
        await self.set_attr_github_token(hash)

        self.create_async_task(self.update_static_readings())

    async def set_restore_now(self, hash, params):
        self.create_async_task(self.restore_backup())

    async def set_attr_github_token(self, hash):
        if self._attr_github_token == "":
            await fhem.readingsSingleUpdate(
                hash, "state", "Please set github_token attribute", 1
            )
            return

        self.gh_token_ready.set()
        github_restore.headers["Authorization"] = f"token {self._attr_github_token}"

    async def restore_backup(self):
        await self.gh_token_ready.wait()
        self.gh_token_ready.clear()

        await fhem.readingsSingleUpdate(self.hash, "state", "Restoring backup", 1)

        # get latest commit from self.directory
        url = f"https://api.github.com/repos/{self.gh_user}/{self.gh_repo}/commits?path={self.directory}"
        commits = await self.github_get(url)
        if not commits:
            await fhem.readingsSingleUpdate(
                self.hash, "state", "Failed to get commits", 1
            )
            return

        commit = commits[0]
        commit_sha = commit["sha"]
        commit_date = commit["commit"]["author"]["date"]
        commit_date = datetime.strptime(commit_date, "%Y-%m-%dT%H:%M:%SZ")
        commit_date = commit_date.strftime("%Y-%m-%d %H:%M:%S")

        # get tree recursively
        url = f"https://api.github.com/repos/{self.gh_user}/{self.gh_repo}/git/trees/{commit_sha}?recursive=1"
        tree = await self.github_get(url)
        if not tree:
            await fhem.readingsSingleUpdate(self.hash, "state", "Failed to get tree", 1)
            return

        # get all files in directory
        backup_files = []
        for entry in tree["tree"]:
            if entry["path"].startswith(self.directory):
                backup_files.append(entry)

        # retrieve all backup files
        for backup_file in backup_files:
            if backup_file["type"] == "blob":
                await self.restore_backup_file(backup_file)

        await fhem.readingsSingleUpdate(
            self.hash,
            "state",
            "Backup restored<br>YOU MUST RESTART FHEM NOW, DO NOT SAVE BEFORE RESTARTING",
            1,
        )

    async def restore_backup_file(self, backup_file):
        # get blob
        url = f"https://api.github.com/repos/{self.gh_user}/{self.gh_repo}/git/blobs/{backup_file['sha']}"
        blob = await self.github_get(url)
        if not blob:
            await fhem.readingsSingleUpdate(self.hash, "state", "Failed to get blob", 1)
            return

        # decode blob
        content = base64.b64decode(blob["content"])

        # create directory, but remove self.directory from path
        file_path = pathlib.Path(backup_file["path"])
        path = file_path.parent
        path = path.relative_to(self.directory)
        path = pathlib.Path(f"{path}")
        path.mkdir(parents=True, exist_ok=True)

        # write file
        path = file_path.relative_to(self.directory)
        path = pathlib.Path(f"{path}")
        path.write_bytes(content)

        # update reading
        await fhem.readingsSingleUpdate(
            self.hash, "state", f"Restored {backup_file['path']}", 1
        )

    async def github_get(self, url):
        ret = None
        async with aiohttp.ClientSession(trust_env=True) as session:
            async with session.get(url, headers=github_restore.headers) as resp:
                if resp.status < 400:
                    ret = await resp.json()
                else:
                    self.logger.error(
                        f"Failed to get {url} with HTTP error {resp.status}"
                    )
        return ret

    async def update_static_readings(self):
        await fhem.readingsSingleUpdateIfChanged(
            self.hash,
            "repository",
            f"<html><a href='{self.url}' target='_blank'>"
            + "Open repository (new tab/window)</a></html>",
            1,
        )



import asyncio
import time
from datetime import datetime
from enum import IntEnum

from .. import fhem, generic
from . import eq3btsmart as eq3


class Mode(IntEnum):
    """Thermostat modes."""

    Unknown = -1
    Closed = 0
    Open = 1
    Auto = 2
    Manual = 3
    Away = 4
    Boost = 5


# TODO set schedules
# TODO set windowOpen, windowOpenTime, eco/comfortTemperature


class eq3bt(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._last_update = 0
        self._mac = None
        self._pin = "0000"
        self.thermostat = None
        self.notification_lock = asyncio.Lock()

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        attr_list = {
            "keep_connected": {
                "default": "on",
                "format": "str",
                "options": "on,off",
                "help": "On...keeps bluetooth low energy connection active which makes commands to be executed immediately",
            },
            "max_retries": {
                "default": 5,
                "format": "int",
                "help": "Maximum retries for connection setup, default=5.",
            },
        }
        await self.set_attr_config(attr_list)

        set_list_conf = {
            "on": {},
            "off": {},
            "desiredTemperature": {
                "args": ["target_temp"],
                "options": "slider,4.5,0.5,30,1",
            },
            "updateStatus": {},
            "boost": {
                "args": ["target_state"],
                "options": "on,off",
                "help": "Activate boost for 300s",
            },
            "mode": {"args": ["target_mode"], "options": "manual,automatic"},
            "eco": {},
            "comfort": {},
            "childlock": {"args": ["target_state"], "options": "on,off"},
            "resetConsumption": {
                "args": ["cons_var"],
                "options": "all,consumption,consumptionToday,consumptionYesterday",
            },
            "temperatureOffset": {
                "args": ["offset"],
                "params": {"offset": {"format": "float"}},
                "options": "slider,-3.5,0.5,3.5,1",
            },
            "windowOpenTemperature": {
                "args": ["temp"],
                "params": {"temp": {"format": "float"}},
                "options": "slider,5,0.5,30,1",
            },
            "windowOpenTime": {
                "args": ["minutes"],
                "params": {"minutes": {"format": "int"}},
                "options": "slider,0,5,60",
            },
            "ecoTemperature": {
                "args": ["temp"],
                "params": {"temp": {"format": "float"}},
                "options": "slider,4.5,0.5,30,1",
            },
            "comfortTemperature": {
                "args": ["temp"],
                "params": {"temp": {"format": "float"}},
                "options": "slider,4.5,0.5,30,1",
            },
        }
        await self.set_set_config(set_list_conf)

        self.hash = hash
        if len(args) < 4:
            return "Usage: define eq3_livingroom fhempy eq3bt MAC PIN"
        self._mac = args[3]
        self.hash["MAC"] = self._mac
        if len(args) > 4:
            self._pin = args[4]
        self.logger.info(f"Define: eq3bt {self._mac}")

        icon = await fhem.AttrVal(self.hash["NAME"], "icon", "noicon")
        if icon == "noicon":
            await fhem.CommandAttr(
                self.hash, self.hash["NAME"] + " icon sani_heating_temp"
            )
        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connecting", 1)

        self.thermostat = FhemThermostat(
            self.logger,
            self.hash,
            self._mac,
            keep_connection=self._attr_keep_connected == "on",
            notification_callback=self.notification_received,
            pin=self._pin,
        )

        self.create_async_task(self.check_online())
        self.create_async_task(self.consumption_rotate())
        return ""

    async def Undefine(self, hash):
        if self.thermostat:
            await self.thermostat.disconnect()
        return await super().Undefine(hash)

    def seconds_till_midnight(self):
        """Get the number of seconds until midnight."""
        n = datetime.now()
        return (
            ((24 - n.hour - 1) * 60 * 60) + ((60 - n.minute - 1) * 60) + (60 - n.second)
        )

    async def consumption_rotate(self):
        """Rotate consumption readings at midnight."""
        while True:
            await asyncio.sleep(self.seconds_till_midnight())
            consumption = float(
                await fhem.ReadingsVal(self.hash["NAME"], "consumptionToday", "0")
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionYesterday", consumption, 1
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionToday", "0", 1
            )

    async def set_resetConsumption(self, hash, params):
        """Reset consumption readings."""
        cons_var = params["cons_var"]
        if cons_var == "all":
            await fhem.readingsSingleUpdateIfChanged(self.hash, "consumption", 0, 1)
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionYesterday", 0, 1
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionToday", 0, 1
            )
        else:
            await fhem.readingsSingleUpdateIfChanged(self.hash, cons_var, 0, 1)

    async def set_attr_keep_connected(self, hash):
        self.thermostat.set_keep_connected(self._attr_keep_connected == "on")

    async def check_online(self):
        """Check if thermostat is online and update readings."""
        await self.thermostat.connect()

        waittime = 300
        if self._attr_keep_connected == "on":
            waittime = 60
        while True:
            try:
                if time.time() - self._last_update > (waittime * 4):
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "update", 1
                    )
                await self.update_all()
            except asyncio.CancelledError:
                self.logger.info("Stopped update loop")
                return
            except asyncio.TimeoutError:
                self.logger.error(f"Timeout on update, retry in {waittime}s")
            except Exception:
                self.logger.exception(f"Failed to update, retry in {waittime}s")
            await asyncio.sleep(waittime)

    async def notification_received(self):
        """Handle notification from thermostat."""
        async with self.notification_lock:
            await self.update_all_readings()

    async def update_all(self):
        """Update all data."""
        self.logger.debug("start update_all")
        await self.thermostat.update_all()

    async def update_all_readings(self):
        """Update all readings."""
        await self.update_readings()
        await self.update_id_readings()
        await self.update_schedule_readings()

    async def update_readings(self):
        """Update readings."""
        old_valve_pos = float(
            await fhem.ReadingsVal(self.hash["NAME"], "valvePosition", "0")
        )
        old_consumption = float(
            await fhem.ReadingsVal(self.hash["NAME"], "consumption", "0")
        )
        old_consumption_today = float(
            await fhem.ReadingsVal(self.hash["NAME"], "consumptionToday", "0")
        )
        await fhem.readingsBeginUpdate(self.hash)
        try:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "battery", self.thermostat.battery
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "boost", self.thermostat.boost
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "childlock", self.thermostat.locked
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "desiredTemperature", self.thermostat.target_temperature
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "ecoTemperature", self.thermostat.eco_temperature
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "temperatureOffset", self.thermostat.temperature_offset
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "comfortTemperature", self.thermostat.comfort_temperature
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "mode", self.thermostat.fhem_mode
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "state", self.thermostat.state
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "completeState", self.thermostat.mode_readable
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "valvePosition", self.thermostat.valve_state
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "awayEnd", self.thermostat.away_end
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "windowOpen", self.thermostat.window_open
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                "windowOpenTemperature",
                self.thermostat.window_open_temperature,
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "windowOpenTime", self.thermostat.window_open_time
            )
            if (time.time() - self._last_update) < 400:
                consumption_diff = (
                    (old_valve_pos + self.thermostat.valve_state)
                    / 2
                    / 100
                    * (time.time() - self._last_update)
                    / 60
                )
            else:
                consumption_diff = 0
            new_consumption = round(old_consumption + consumption_diff, 2)
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "consumption", new_consumption
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                "consumptionToday",
                round(old_consumption_today + consumption_diff, 2),
            )
        except Exception:
            self.logger.exception("Failed to update_readings")
            raise
        finally:
            await fhem.readingsEndUpdate(self.hash, 1)
        self._last_update = time.time()

    async def update_id_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "firmware", self.thermostat.firmware_version
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "serialNumber", self.thermostat.device_serial
        )
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_schedule_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        for day in list(self.thermostat.schedule.keys()):
            reading = f"schedule_{day}_1"
            if self.thermostat.schedule[day].base_temp == 0 or isinstance(
                self.thermostat.schedule[day].next_change_at, int
            ):
                await fhem.readingsBulkUpdateIfChanged(self.hash, reading, "-")
                last_change = "00:00"
            else:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash,
                    reading,
                    f"00:00 - {self.thermostat.schedule[day].next_change_at.strftime('%H:%M')}: {self.thermostat.schedule[day].base_temp}",
                )
                last_change = self.thermostat.schedule[day].next_change_at.strftime(
                    "%H:%M"
                )
            last_schedule = False
            for h in range(0, 6):
                reading = f"schedule_{day}_{h+2}"
                if (
                    h == 6
                    or self.thermostat.schedule[day].hours[h].target_temp == 0
                    or isinstance(
                        self.thermostat.schedule[day].hours[h].next_change_at, int
                    )
                    or last_schedule
                ):
                    if last_schedule:
                        await fhem.readingsBulkUpdateIfChanged(self.hash, reading, "-")
                    else:
                        value = f"{last_change} - 00:00: {self.thermostat.schedule[day].base_temp}"
                        await fhem.readingsBulkUpdateIfChanged(
                            self.hash, reading, value
                        )
                    last_schedule = True
                else:
                    value = f"{last_change} - {self.thermostat.schedule[day].hours[h].next_change_at.strftime('%H:%M')}: {self.thermostat.schedule[day].hours[h].target_temp}"
                    last_change = (
                        self.thermostat.schedule[day]
                        .hours[h]
                        .next_change_at.strftime("%H:%M")
                    )
                    await fhem.readingsBulkUpdateIfChanged(self.hash, reading, value)
        await fhem.readingsEndUpdate(self.hash, 1)

    async def set_and_update(self, fct):
        await fct
        await self.update_readings()

    def string_to_seconds(self, timestr):
        ftr = [3600, 60, 1]
        return sum([a * b for a, b in zip(ftr, map(int, timestr.split(":")))])

    # SET Functions BEGIN
    async def set_on(self, hash, params):
        self.create_async_task(
            self.set_and_update(self.thermostat.set_target_temperature(30))
        )

    async def set_off(self, hash, params):
        self.create_async_task(
            self.set_and_update(self.thermostat.set_target_temperature(4.5))
        )

    async def set_desiredTemperature(self, hash, params):
        temp = float(params["target_temp"])
        self.create_async_task(
            self.set_and_update(self.thermostat.set_target_temperature(temp))
        )

    async def set_ecoTemperature(self, hash, params):
        comfort_temp = float(
            await fhem.ReadingsVal(self.hash["NAME"], "comfortTemperature", "17.0")
        )
        self.create_async_task(
            self.set_and_update(
                self.thermostat.set_temperature_presets(
                    comfort_temp,
                    params["temp"],
                )
            )
        )

    async def set_comfortTemperature(self, hash, params):
        eco_temp = float(
            await fhem.ReadingsVal(self.hash["NAME"], "ecoTemperature", "15.0")
        )
        self.create_async_task(
            self.set_and_update(
                self.thermostat.set_temperature_presets(params["temp"], eco_temp)
            )
        )

    async def set_temperatureOffset(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                self.thermostat.set_temperature_offset(params["offset"])
            )
        )

    async def set_windowOpenTemperature(self, hash, params):
        duration = await fhem.ReadingsVal(
            self.hash["NAME"], "windowOpenTime", "0:15:00"
        )
        duration_sec = self.string_to_seconds(duration)
        self.create_async_task(
            self.set_and_update(
                self.thermostat.set_window_open_config(
                    params["temp"],
                    duration_sec,
                )
            )
        )

    async def set_windowOpenTime(self, hash, params):
        temp = await fhem.ReadingsVal(self.hash["NAME"], "windowOpenTemperature", "5")
        temp = float(temp)
        self.create_async_task(
            self.set_and_update(
                self.thermostat.set_window_open_config(temp, params["minutes"] * 60)
            )
        )

    async def set_updateStatus(self, hash, params):
        self.create_async_task(self.update_all())

    async def set_boost(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                self.thermostat.set_boost(params["target_state"] == "on")
            )
        )

    async def set_mode(self, hash, params):
        target_mode = params["target_mode"]
        if target_mode == "automatic":
            target_mode = eq3.Mode.Auto
        else:
            target_mode = eq3.Mode.Manual
        self.create_async_task(
            self.set_and_update(self.thermostat.set_fhem_mode(target_mode))
        )

    async def set_eco(self, hash, params):
        self.create_async_task(self.set_and_update(self.thermostat.activate_eco()))

    async def set_comfort(self, hash, params):
        self.create_async_task(self.set_and_update(self.thermostat.activate_comfort()))

    async def set_childlock(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                self.thermostat.set_locked(params["target_state"] == "on")
            )
        )

    # SET Functions END


class FhemThermostat(eq3.Thermostat):
    def __init__(self, logger, hash, mac, keep_connection, notification_callback, pin):
        self.logger = logger
        self._keep_conection = keep_connection
        super(FhemThermostat, self).__init__(
            logger,
            hash,
            mac,
            keep_connection=self._keep_conection,
            notification_callback=notification_callback,
            pin=pin,
        )

    async def disconnect(self):
        await super().disconnect()

    async def update_all(self):
        await super().update()
        await asyncio.sleep(3)
        await super().query_id()
        await asyncio.sleep(3)
        for day in range(0, 7):
            await super().query_schedule(day)
            await asyncio.sleep(1)

    async def set_temperature_presets(self, comfort_temp, eco_temp):
        await super().temperature_presets(comfort_temp, eco_temp)

    async def set_temperature_offset(self, temp):
        await super().set_temperature_offset(temp)

    async def set_window_open_config(self, temperature, duration):
        await super().window_open_config(temperature, duration)

    async def set_target_temperature(self, temp):
        await super().set_target_temperature(temp)

    async def set_boost(self, state):
        await super().set_boost(state)

    async def set_locked(self, state):
        await super().set_locked(state)

    async def set_fhem_mode(self, mode):
        await super().set_mode(mode)

    @property
    def fhem_mode(self):
        if self._mode == Mode.Boost:
            return "boost"
        elif self._mode == Mode.Away:
            return "away"
        elif self._mode == Mode.Closed:
            return "manual"
        elif self._mode == Mode.Open:
            return "manual"
        elif self._mode == Mode.Manual:
            return "manual"
        elif self._mode == Mode.Auto:
            return "automatic"

    @property
    def state(self):
        if self._mode == Mode.Boost:
            return "boost"
        elif self._mode == Mode.Away:
            return "away"
        elif self._mode == Mode.Closed:
            return "off"
        elif self._mode == Mode.Open:
            return "on"
        elif self._mode == Mode.Manual:
            return "manual"
        elif self._mode == Mode.Auto:
            return "automatic"

    @property
    def battery(self):
        if self.low_battery:
            return "low"
        return "ok"



import asyncio
import logging
import re

from pyit600.exceptions import IT600AuthenticationError, IT600ConnectionError
from pyit600.gateway_singleton import IT600GatewaySingleton

from .. import fhem, generic

_LOGGER = logging.getLogger(__name__)


class pyit600(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        self.attr_config = {
            "update_interval": {
                "default": 60,
                "format": "int",
                "help": "Readings update intervall in seconds (default 60s).",
            },
            "update_readings": {
                "default": "always",
                "options": "always,onchange",
                "help": "Update readings only on value change or always (default always).",
            },
            "update_min_interval": {
                "default": 0,
                "format": "int",
                "help": "Update by interval in combination with onchange. Add multiplier (default 0 -> off, e.g. 5 -> 5x update_interval).",
            },
            "readings_device_id": {
                "default": "off",
                "options": "off,on",
                "help": "Create readings with device ID, also if device name is available (default off).",
            },
            "readings_cmpl_dev_str": {
                "default": "on",
                "options": "on,off",
                "help": "Create readings with all values of device (default on).",
            },
        }
        await self.set_attr_config(self.attr_config)

        if len(args) != 5:
            return "Usage: define my_salus_device fhempy pyit600 host euid"
        self.host = args[3]
        self.euid = args[4]
        self.debug = 1
        self.min_interval_timer = 1
        _LOGGER.debug(f"host: %s, euid: %s", self.host, self.euid)
        self.create_async_task(self.start_login())
        await fhem.readingsSingleUpdate(self.hash, "state", "connecting", 1)

    async def start_login(self):
        async with IT600GatewaySingleton.get_instance(
            host=self.host, euid=self.euid, debug=self.debug
        ) as self.gateway:
            try:
                _LOGGER.debug(f"start login")
                await self.gateway.connect()
                await self.gateway.poll_status(send_callback=False)
                await self.prepare_set_commands()
                await self.update_readings()
            except IT600ConnectionError:
                await fhem.readingsSingleUpdate(
                    self.hash, "state", "Connection error", 1
                )
                _LOGGER.info(f"Connection error: check if you have specified gateway")
            except IT600AuthenticationError:
                await fhem.readingsSingleUpdate(
                    self.hash, "state", "Authentication error", 1
                )
                _LOGGER.info(
                    f"Authentication error: check if you have specified gateway EUID is correctly."
                )

    async def prepare_set_commands(self):
        self.set_config = {
            "target_temperature": {
                "args": ["climate_device_id", "temperature"],
                "params": {
                    "climate_device_id": {"format": "str"},
                    "temperature": {"default": 23, "format": "float"},
                },
                "help": (
                    "Parameters: deviceID Temperature <br>"
                    "e.g. set my_salus_device target_temperature 001e5e09020770e3 23<br>"
                ),
            }
        }
        await self.set_set_config(self.set_config)

    async def set_target_temperature(self, hash, params):
        self.create_async_task(
            self.gateway.set_climate_device_temperature(
                params["climate_device_id"], params["temperature"]
            )
        )

    async def update_readings(self):
        while True:
            _LOGGER.debug(f"update reading")
            await self.update_readings_once()
            await asyncio.sleep(self._attr_update_interval)

    async def update_readings_once(self):
        for remaining_attempts in reversed(range(60)):
            try:
                await self.gateway.poll_status(send_callback=False)
                break
            except Exception as e:
                if remaining_attempts == 0:
                    await fhem.readingsSingleUpdate(
                        self.hash, "state", "Connection timeout: retrying stopped", 1
                    )
                    raise e
                else:
                    _LOGGER.info(
                        f"Connection timeout - remaining attempts: %d",
                        remaining_attempts,
                    )
                    await fhem.readingsSingleUpdate(
                        self.hash, "state", "Connection timeout: retrying ...", 1
                    )
                    await asyncio.sleep(self._attr_update_interval)

        try:
            climate_devices = self.gateway.get_climate_devices()

            if not climate_devices:
                await fhem.readingsSingleUpdate(
                    self.hash, "state", "No device found", 1
                )
            else:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", "online", 1
                )

                if (self._attr_update_readings == "always") or (
                    (self._attr_update_min_interval > 0)
                    and (self.min_interval_timer >= self._attr_update_min_interval)
                ):

                    self.min_interval_timer = 1

                    for climate_device_id in climate_devices:

                        if self._attr_readings_cmpl_dev_str == "on":
                            await fhem.readingsSingleUpdate(
                                self.hash,
                                climate_device_id,
                                climate_devices.get(climate_device_id),
                                1,
                            )

                        try:
                            name = re.search(
                                "name=" "(.+?)" ",",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)
                            name = name.replace(" ", "")
                            name = name.replace("'", "")
                            curtemp = re.search(
                                "current_temperature=(.+?),",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)

                            if (
                                name == "unknown"
                                or self._attr_readings_device_id == "on"
                            ):
                                await fhem.readingsSingleUpdate(
                                    self.hash,
                                    climate_device_id + "_curtemp",
                                    curtemp,
                                    1,
                                )
                            else:
                                await fhem.readingsSingleUpdate(
                                    self.hash, name + "_curtemp", curtemp, 1
                                )
                        except AttributeError:
                            _LOGGER.info(
                                f"Error on reading device name or current temperature!"
                            )

                        try:
                            name = re.search(
                                "name=" "(.+?)" ",",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)
                            name = name.replace(" ", "")
                            name = name.replace("'", "")
                            targtemp = re.search(
                                "target_temperature=(.+?),",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)

                            if (
                                name == "unknown"
                                or self._attr_readings_device_id == "on"
                            ):
                                await fhem.readingsSingleUpdate(
                                    self.hash,
                                    climate_device_id + "_targtemp",
                                    targtemp,
                                    1,
                                )
                            else:
                                await fhem.readingsSingleUpdate(
                                    self.hash, name + "_targtemp", targtemp, 1
                                )
                        except AttributeError:
                            _LOGGER.info(
                                f"Error on reading device name or target temperature!"
                            )

                        try:
                            name = re.search(
                                "name=" "(.+?)" ",",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)
                            name = name.replace(" ", "")
                            name = name.replace("'", "")
                            hvac_act = re.search(
                                "hvac_action=(.+?),",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)

                            if (
                                name == "unknown"
                                or self._attr_readings_device_id == "on"
                            ):
                                await fhem.readingsSingleUpdate(
                                    self.hash,
                                    climate_device_id + "_hvac_action",
                                    hvac_act,
                                    1,
                                )
                            else:
                                await fhem.readingsSingleUpdate(
                                    self.hash, name + "_hvac_action", hvac_act, 1
                                )
                        except AttributeError:
                            _LOGGER.info(
                                f"Error on reading device name or hvac_action!"
                            )							

                else:
                    self.min_interval_timer += 1

                    for climate_device_id in climate_devices:
                        if self._attr_readings_cmpl_dev_str == "on":
                            await fhem.readingsSingleUpdateIfChanged(
                                self.hash,
                                climate_device_id,
                                climate_devices.get(climate_device_id),
                                1,
                            )

                        try:
                            name = re.search(
                                "name=" "(.+?)" ",",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)
                            name = name.replace(" ", "")
                            name = name.replace("'", "")
                            curtemp = re.search(
                                "current_temperature=(.+?),",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)

                            if (
                                name == "unknown"
                                or self._attr_readings_device_id == "on"
                            ):
                                await fhem.readingsSingleUpdateIfChanged(
                                    self.hash,
                                    climate_device_id + "_curtemp",
                                    curtemp,
                                    1,
                                )
                            else:
                                await fhem.readingsSingleUpdateIfChanged(
                                    self.hash, name + "_curtemp", curtemp, 1
                                )
                        except AttributeError:
                            _LOGGER.info(
                                f"Error on reading device name or current temperature!"
                            )

                        try:
                            name = re.search(
                                "name=" "(.+?)" ",",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)
                            name = name.replace(" ", "")
                            name = name.replace("'", "")
                            targtemp = re.search(
                                "target_temperature=(.+?),",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)

                            if (
                                name == "unknown"
                                or self._attr_readings_device_id == "on"
                            ):
                                await fhem.readingsSingleUpdateIfChanged(
                                    self.hash,
                                    climate_device_id + "_targtemp",
                                    targtemp,
                                    1,
                                )
                            else:
                                await fhem.readingsSingleUpdateIfChanged(
                                    self.hash, name + "_targtemp", targtemp, 1
                                )
                        except AttributeError:
                            _LOGGER.info(
                                f"Error on reading device name or target temperature!"
                            )

                        try:
                            name = re.search(
                                "name=" "(.+?)" ",",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)
                            name = name.replace(" ", "")
                            name = name.replace("'", "")
                            hvac_act = re.search(
                                "hvac_action=(.+?),",
                                str(climate_devices.get(climate_device_id)),
                            ).group(1)

                            if (
                                name == "unknown"
                                or self._attr_readings_device_id == "on"
                            ):
                                await fhem.readingsSingleUpdateIfChanged(
                                    self.hash,
                                    climate_device_id + "_hvac_action",
                                    hvac_act,
                                    1,
                                )
                            else:
                                await fhem.readingsSingleUpdateIfChanged(
                                    self.hash, name + "_hvac_action", hvac_act, 1
                                )
                        except AttributeError:
                            _LOGGER.info(
                                f"Error on reading device name or hvac_action!"
                            )
        except Exception:
            self.logger.exception("Failed to update readings")



import asyncio

import httpx
import pyprusalink

from fhempy.lib.utils import flatten_json

from .. import fhem, generic


class prusalink(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 6:
            return "Usage: define prusa_mk4 fhempy prusalink URL USERNAME PASSWORD"

        await self.set_icon("3d_printer")

        self.url = args[3]
        self.username = args[4]
        self.password = args[5]

        self.create_async_task(self.update_loop())

    async def retrieve_data(self):
        data = {}
        try:
            data["status"] = await self.pl.get_status()
        except Exception as ex:
            self.logger.exception("Failed to update")
            await fhem.readingsSingleUpdate(
                self.hash,
                "state",
                f"failed: {ex}",
                1,
            )
        return data

    async def update_loop(self):
        async with httpx.AsyncClient() as session:
            self.pl = pyprusalink.PrusaLink(
                session, self.url, self.username, self.password
            )
            while True:
                data = await self.retrieve_data()
                flat_json = flatten_json(data)
                await fhem.readingsBeginUpdate(self.hash)
                try:
                    for name in flat_json:
                        await fhem.readingsBulkUpdateIfChanged(
                            self.hash, name, flat_json[name]
                        )
                    await fhem.readingsBulkUpdateIfChanged(
                        self.hash, "state", flat_json["status_printer_state"]
                    )
                except Exception:
                    self.logger.exception("Failed to update readings")
                await fhem.readingsEndUpdate(self.hash, 1)

                await asyncio.sleep(30)



import asyncio

import goodwe as gw
from goodwe.inverter import OperationMode

from .. import fhem, generic, utils


class goodwe(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {
            "interval": {
                "default": 300,
                "format": "int",
                "help": "Change interval in seconds, default is 300s.",
            }
        }
        await self.set_attr_config(attr_config)

        set_config = {
            "operation_mode": {
                "args": ["mode"],
                "options": "general,off_grid,backup,eco,peak_shaving,eco_charge,eco_discharge",
                "help": "Change operations mode of the inverter.",
            }
        }
        await self.set_set_config(set_config)

        if len(args) != 4:
            return "Usage: define inverter fhempy goodwe IP"

        self.ip = args[3]

        self.create_async_task(self.update_loop())

    async def set_operation_mode(self, hash, params):
        if params["mode"] == "general":
            set_mode = OperationMode.GENERAL
        elif params["mode"] == "off_grid":
            set_mode = OperationMode.OFF_GRID
        elif params["mode"] == "backup":
            set_mode = OperationMode.BACKUP
        elif params["mode"] == "eco":
            set_mode = OperationMode.ECO
        elif params["mode"] == "peak_shaving":
            set_mode = OperationMode.PEAK_SHAVING
        elif params["mode"] == "eco_charge":
            set_mode = OperationMode.ECO_CHARGE
        elif params["mode"] == "eco_discharge":
            set_mode = OperationMode.ECO_DISCHARGE
        else:
            set_mode = OperationMode.GENERAL
        self.create_async_task(
            self.inverter.set_operation_mode(
                set_mode
            )
        )

    def get_enum_name(self, enum_value):
        for name, value in OperationMode.__members__.items():
            if value == enum_value:
                return name

    async def update_loop(self):
        self.inverter = None
        while self.inverter is None:
            self.inverter = await gw.connect(self.ip)
            await asyncio.sleep(30)

        while True:
            try:
                runtime_data = await self.inverter.read_runtime_data()
                await self.handle_data(runtime_data)

                operation_mode = await self.inverter.get_operation_mode()
                await fhem.readingsSingleUpdate(self.hash, "operation_mode", self.get_enum_name(operation_mode))
            except Exception:
                await fhem.readingsSingleUpdate(self.hash, "state", "error", 1)
            await asyncio.sleep(self._attr_interval)

    async def handle_data(self, runtime_data):
        await fhem.readingsBeginUpdate(self.hash)
        try:
            for sensor in self.inverter.sensors():
                if sensor.id_ in runtime_data:
                    reading = utils.gen_reading_name(sensor.name)
                    if sensor.unit:
                        reading += "_" + sensor.unit
                    await fhem.readingsBulkUpdate(
                        self.hash,
                        reading,
                        runtime_data[sensor.id_],
                    )
            await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "connected")
        except Exception:
            self.logger.exception("Failed to update readings")
        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio

from .. import fhem, generic
from alphaess.alphaess import alphaess
from datetime import date, timedelta

class alphaesscloud(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.appID = None
        self.appSecret = None
        self.serial = None
        self.ESSunit = None
        self.client = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        attr_config = {
            "interval": {
                "default": 60,
                "format": "int",
                "help": "Change interval, default is 60.",
            }
        }
        await self.set_attr_config(attr_config)

        if len(args) !=6:
            return "Usage: define myAlphaESSCloud fhempy alphaesscloud appID appSecret serial"
        
        self.appID = args[3]
        self.appSecret = args[4]
        self.serial = args[5]
        
        self.create_async_task(self.setup_connection())

    async def setup_connection(self):
            """
            Sets up the connection to the AlphaESS cloud and retrieves necessary information.

            This method connects to the AlphaESS cloud using the provided app ID and app secret.
            It retrieves the list of ESS units associated with the account and checks if the
            specified serial number matches any of the units. If a match is found, it updates
            the state to "connected" and updates the readings from the unit's item list.

            If the serial number does not match any of the units, it updates the state to
            "ESS unit not found please check serial number".

            Additionally, it retrieves the discharging and charging configuration information
            for the ESS unit and updates the readings accordingly.

            Finally, it kicks off the routine to update the data from the cloud continuously.

            Returns:
                None
            """
               
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connecting",1)

            self.client: alphaess = alphaess(self.appID,self.appSecret)
            ESSList = await self.client.getESSList()
            if ESSList != None:
                self.logger.error(f"Retrieved: {len(ESSList)} ESS units")
            else:
                await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "no ESS units found ",1)

            for unit in ESSList:
                if "sysSn" in unit:
                    if unit["sysSn"] == self.serial:
                        self.ESSunit = unit
                        
                        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connected",1)
                        await self.update_readings_from_itemlist(unit)
                        break

            if self.ESSunit == None:
                await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "ESS unit not found please check serial number",1)
                return
            
            dischargingInfo = await self.client.getDisChargeConfigInfo(self.serial)
            if dischargingInfo != None:
                await self.update_readings_from_itemlist(dischargingInfo)
            else:
                self.logger.error("No discharging info found")

            chargingInfo = await self.client.getChargeConfigInfo(self.serial)
            if chargingInfo != None:
                await self.update_readings_from_itemlist(chargingInfo)
            else:
                self.logger.error("No charging info found")

            await self.update_data_from_cloud()



      

    

    async def update_data_from_cloud(self):
            """
            Updates the data from the cloud for the AlphaESS device.

            This method continuously retrieves data from the cloud for the specified AlphaESS device.
            It retrieves the sum data and last power data from the cloud and updates the readings in FHEM.
            If an exception occurs during the update process, the state reading is updated to indicate the failure.

            
            Returns:
            None
            """
            while True:
                try:
                    sumData = await self.client.getSumDataForCustomer(self.serial)
                    if sumData != None:
                        await self.update_readings_from_itemlist(sumData)

                    lastPower = await self.client.getLastPowerData(self.serial)
                    if lastPower != None:
                        await self.update_readings_from_itemlist(lastPower)

                except Exception:
                    await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "data update failed", 1)

                if self._attr_interval > 10:
                    await asyncio.sleep(self._attr_interval)
                else:
                    await asyncio.sleep(10)

    async def update_readings_from_itemlist(self, itemlist):
            """
            Updates the readings from the given itemlist via a bluk update.

            Args:
                itemlist (dict): A dictionary containing the readings to be updated.

            Returns:
                None
            """
            await fhem.readingsBeginUpdate(self.hash)
            for key, value in itemlist.items():
                if isinstance(value, dict):
                    for sub_key, sub_value in value.items():
                        await fhem.readingsBulkUpdate(self.hash, sub_key, sub_value)
                else:
                        await fhem.readingsBulkUpdate(self.hash, key, value)
            await fhem.readingsEndUpdate(self.hash, 1)


    



import asyncio
from datetime import datetime

import RPi.GPIO as GPIO
from luma.core.interface.serial import noop, spi
from luma.core.render import canvas
from luma.core.virtual import viewport
from luma.led_matrix.device import max7219
from PIL import ImageFont

from .. import fhem, generic, utils


class piclock(generic.FhemModule):

    pm_font = ImageFont.truetype(utils.get_fhempy_root() + "/piclock/clockfont.ttf", 8)
    icons = {
        "rain": [0x1C, 0x7E, 0xFF, 0xFF, 0x7E, 0x8, 0x52, 0x84],
        "nt_rain": [0x1C, 0x7E, 0xFF, 0xFF, 0x7E, 0x8, 0x52, 0x84],
        "chancerain": [0x1C, 0x7E, 0xFF, 0xFF, 0x7E, 0x8, 0x52, 0x84],
        "nt_chancerain": [0x1C, 0x7E, 0xFF, 0xFF, 0x7E, 0x8, 0x52, 0x84],
        "clear": [0x81, 0x5A, 0x3C, 0x7E, 0x7E, 0x3C, 0x5A, 0x81],
        "nt_clear": [0x30, 0x18, 0xC, 0xE, 0xE, 0xC, 0x18, 0x30],
        "snow": [0x42, 0x24, 0x99, 0x7E, 0x99, 0x24, 0x42, 0x00],
        "sleet": [0x1C, 0x7E, 0xFF, 0xFF, 0x7E, 0x89, 0x20, 0x85],
        "wind": [0x0, 0x6, 0x1, 0xFE, 0x0, 0xFE, 0x1, 0x6],
        "fog": [0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA],
        "cloudy": [0x1C, 0x7E, 0xFF, 0xFF, 0x7E, 0x0, 0x0, 0x0],
        "nt_cloudy": [0x1C, 0x7E, 0xFF, 0xFF, 0x7E, 0x0, 0x0, 0x0],
        "mostlycloudy": [0x1C, 0x7E, 0xFF, 0xFF, 0x7E, 0x0, 0x0, 0x0],
        "nt_mostlycloudy": [0x1C, 0x7E, 0xFF, 0xFF, 0x7E, 0x0, 0x0, 0x0],
        "partlycloudy": [0x0, 0x1C, 0x62, 0x81, 0x81, 0x7E, 0x0, 0x0],
        "nt_partlycloudy": [0x0, 0x1C, 0x62, 0x81, 0x81, 0x7E, 0x0, 0x0],
        "tstorm": [0x1C, 0x62, 0x81, 0x81, 0x7E, 0x18, 0xC, 0x10],
        "notused_nt_partlycloudy": [0xFF, 0xF3, 0x9D, 0x7E, 0x7E, 0x81, 0xFF, 0xFF],
    }

    def __init__(self, logger):
        super().__init__(logger)

        self.temp_dev = None
        self.temp_reading = None
        self.weather_dev = None
        self.weather_reading = None
        self.buttons = []
        self._notification_inactive = asyncio.Event()

        # create seven segment device
        serial = spi(port=0, device=0, gpio=noop())
        self.matrix_device = max7219(
            serial, cascaded=4, block_orientation=-90, rotate=0
        )

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {
            "display_time_duration": {
                "default": 20,
                "format": "int",
                "help": "Duration of time display in seconds (default 20s).",
            },
            "display_weather_duration": {
                "default": 5,
                "format": "int",
                "help": "Duration of time display in seconds (default 5s).",
            },
        }
        await self.set_attr_config(attr_config)

        set_config = {
            "notification": {
                "args": ["text"],
                "help": "Display a notification on piclock.",
            },
            "brightness": {
                "args": ["contrast"],
                "params": {"contrast": {"format": "int"}},
                "options": "slider,0,1,255,0",
            },
        }
        await self.set_set_config(set_config)

        if len(args) > 6 or len(args) < 3:
            return (
                "Usage: define piclock fhempy piclock "
                "DEVICE:TEMP_READING DEVICE:WEATHER_READING GPIOID_LIST_OF_BUTTONS"
            )

        if len(args) >= 4:
            self.temp_dev, self.temp_reading = args[3].split(":")
        if len(args) >= 5:
            self.weather_dev, self.weather_reading = args[4].split(":")
        if len(args) == 6:
            self.buttons = args[5].split(",")

        # init buttons
        GPIO.setmode(GPIO.BCM)
        for input in self.buttons:  # [4, 17, 27, 22, 18]
            GPIO.setup(input, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            GPIO.add_event_detect(
                input, GPIO.RISING, callback=self.button_pressed, bouncetime=300
            )

        await fhem.readingsSingleUpdate(self.hash, "state", "active", 1)
        self.change_contrast(
            int(await fhem.ReadingsVal(self.hash["NAME"], "brightness", "100"))
        )

        self.create_async_task(self.run_piclock())

    async def set_notification(self, hash, params):
        self.create_async_task(self.show_notification(params["text"]))

    async def set_brightness(self, hash, params):
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, "brightness", params["contrast"], 1
        )
        self.change_contrast(params["contrast"])

    async def run_piclock(self):
        await self.display_text("PiClock loading...")
        await asyncio.sleep(5)
        await self.run_piclock_mainloop()

    async def run_piclock_mainloop(self):
        while True:
            try:
                # show time
                self._current_time_task = self.create_async_task(
                    self.show_current_time_loop()
                )

                # show weather
                if self.temp_dev is not None:
                    await asyncio.sleep(self._attr_display_time_duration)
                    self._current_time_task.cancel()

                    self._current_weather_task = self.create_async_task(
                        self.show_current_weather()
                    )
                    await asyncio.sleep(self._attr_display_weather_duration)
                    self._current_weather_task.cancel()
                else:
                    # no weather data, show only time
                    return

            except asyncio.CancelledError:
                await self._notification_inactive.wait()
            except Exception:
                self.logger.exception("Failed run_piclock_mainloop")

    async def show_notification(self, notification):
        self._notification_inactive.clear()
        if self._current_time_task.done() is False:
            self._current_time_task.cancel()
        if self._current_weather_task.done() is False:
            self._current_weather_task.cancel()
        await self.display_text(notification)
        self._notification_inactive.set()

    async def show_current_time_loop(self):
        try:
            while True:
                now = datetime.now()
                await self.display_text(now.strftime("%H %M"))
                next_change = 60 - datetime.now().second
                await asyncio.sleep(next_change)
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger.exception("Failed on show_current_time_loop")

    async def show_current_weather(self):
        weather = None

        if self.temp_dev is not None:
            temp = await fhem.ReadingsVal(
                self.temp_dev, self.temp_reading, "No temperature data available"
            )
        if self.weather_dev is not None:
            weather = await fhem.ReadingsVal(self.weather_dev, self.weather_reading, "")
            if weather == "":
                weather = None

        if self.temp_dev is not None:
            await self.display_text(temp + "°", weather)

    async def display_text(self, msg, iconid=None):
        await fhem.readingsSingleUpdate(self.hash, "display", msg, 1)

        w = self.text_size(msg) + self.icon_size(iconid)
        x = self.matrix_device.width
        startx = 0

        virtual = viewport(
            self.matrix_device, width=w + x, height=self.matrix_device.height
        )
        with canvas(virtual) as draw:
            if w < x:
                startx = (x - w) / 2 + 1
            if iconid is not None:
                startx += self.display_icon(draw, startx, iconid)

            draw.text((startx, 0), msg, fill="white", font=piclock.pm_font)

        # sroll text
        if w > x:
            await asyncio.sleep(1.5)
            for offset in range(w - x):
                virtual.set_position((offset, 0))
                await asyncio.sleep(0.04)
            await asyncio.sleep(1.8)

    def change_contrast(self, contrast):
        # change contrast
        self.matrix_device.contrast(contrast)

    def text_size(self, msg):
        w = 0
        with canvas(self.matrix_device) as draw:
            w = draw.textlength(msg, font=piclock.pm_font)
        return int(w)

    def icon_size(self, iconid):
        if iconid is None:
            return 0

        if iconid in piclock.icons:
            return 9
        else:
            return 0

    def display_icon(self, draw, startx, iconid):
        iconid_small = iconid.replace(" ", "").lower()
        if iconid_small in piclock.icons:
            x, y = (0, startx)
            for byte in piclock.icons[iconid_small]:
                for j in range(7, -1, -1):
                    if byte & 0x01 > 0:
                        draw.point((y + j, x), fill="white")
                    byte >>= 1
                x += 1
            return 9
        else:
            self.logger.error("icon not found: " + iconid_small)
            return 0

    def button_pressed(self, channel):
        self.create_async_task(self.async_button_pressed(channel))

    async def async_button_pressed(self, button):
        if button > 0:
            text = await fhem.AttrVal(
                self.hash["NAME"], "button_" + str(button) + "_display"
            )
            cmd = await fhem.AttrVal(
                self.hash["NAME"], "button_" + str(button) + "_cmd"
            )

            await self.show_notification(text)
            await fhem.sendCommandName(self.hash["NAME"], cmd)



# Based on ReachView code from Egor Fedorov (egor.fedorov@emlid.com)
# Updated for Python 3.6.8 on a Raspberry  Pi


import subprocess
import time

import pexpect


class Bluetoothctl:
    """A wrapper for bluetoothctl utility."""

    def __init__(self, logger):
        self.logger = logger
        # try to unblock bluetooth with rfkill, it doesn't matter if it fails
        try:
            subprocess.check_output(
                "rfkill unblock bluetooth",
                shell=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        except Exception as e:
            self.logger.debug("rfkill is not present")
        self.process = pexpect.spawn("bluetoothctl", echo=False, encoding="utf-8")

    def stop(self):
        """Stop bluetoothctl utility only if it is running."""
        if self.process.isalive():
            self.process.terminate(force=True)

    def send(self, command, pause=0):
        self.process.send(f"{command}\n")
        time.sleep(pause)
        if self.process.expect(["#", pexpect.EOF]):
            raise Exception(f"failed after {command}")

    def get_output(self, *args, **kwargs):
        """Run a command in bluetoothctl prompt, return output as a list of lines."""
        self.send(*args, **kwargs)
        return self.process.before.split("\r\n")

    def exit(self):
        """Exit bluetoothctl prompt."""
        self.send("exit")

    def power_on(self):
        """Power on."""
        try:
            self.send("power on")
        except Exception as e:
            self.logger.error(e)

    def power_off(self):
        """Power off."""
        try:
            self.send("power off")
        except Exception as e:
            self.logger.error(e)

    def default_agent(self):
        """Default agent."""
        try:
            self.send("default-agent")
        except Exception as e:
            self.logger.error(e)

    def agent_on(self):
        """Agent on."""
        try:
            self.send("agent on")
        except Exception as e:
            self.logger.error(e)

    def start_scan(self):
        """Start bluetooth scanning process."""
        try:
            self.send("scan on")
        except Exception as e:
            self.logger.error(e)

    def stop_scan(self):
        """Stop bluetooth scanning process."""
        try:
            self.send("scan off")
        except Exception as e:
            self.logger.error(e)

    def make_discoverable(self):
        """Make device discoverable."""
        try:
            self.send("discoverable on")
        except Exception as e:
            self.logger.error(e)

    def parse_device_info(self, info_string):
        """Parse a string corresponding to a device."""
        device = {}
        block_list = ["[\x1b[0;", "removed"]
        if not any(keyword in info_string for keyword in block_list):
            try:
                device_position = info_string.index("Device")
            except ValueError:
                pass
            else:
                if device_position > -1:
                    attribute_list = info_string[device_position:].split(" ", 2)
                    device = {
                        "mac_address": attribute_list[1],
                        "name": attribute_list[2],
                    }
        return device

    def get_available_devices(self):
        """Return a list of tuples of paired and discoverable devices."""
        available_devices = []
        try:
            out = self.get_output("devices")
        except Exception as e:
            self.logger.error(e)
        else:
            for line in out:
                device = self.parse_device_info(line)
                if device:
                    available_devices.append(device)
        return available_devices

    def get_paired_devices(self):
        """Return a list of tuples of paired devices."""
        paired_devices = []
        try:
            out = self.get_output("paired-devices")
            # add output of devices Paired command
            out += self.get_output("devices Paired")
        except Exception as e:
            self.logger.error(e)
        else:
            for line in out:
                device = self.parse_device_info(line)
                if device:
                    paired_devices.append(device)
        return paired_devices

    def get_discoverable_devices(self):
        """Filter paired devices out of available."""
        available = self.get_available_devices()
        paired = self.get_paired_devices()
        return [d for d in available if d not in paired]

    def get_device_info(self, mac_address):
        """Get device info by mac address."""
        try:
            out = self.get_output(f"info {mac_address}")
        except Exception as e:
            self.logger.error(e)
            return False
        else:
            return out

    def pair(self, mac_address, pin="0000"):
        """Try to pair with a device by mac address."""
        try:
            self.send(f"pair {mac_address}", 4)
        except Exception as e:
            self.logger.error(e)
            return PairingState.FAILED
        else:
            try:
                res = self.process.expect(
                    [
                        "Pairing successful",
                        "Request passkey",
                        "Failed to pair",
                        pexpect.TIMEOUT,
                    ],
                    timeout=60,
                )
                if res == 1:
                    self.send(pin, 4)
                    res = self.process.expect(
                        [
                            "Pairing successful",
                            "Failed to pair",
                            "AuthenticationFailed",
                            pexpect.TIMEOUT,
                        ],
                        timeout=60,
                    )
                    if res == 2:
                        return PairingState.WRONG_PIN
                    elif res == 3:
                        return PairingState.TIMEOUT
                    elif res == 1:
                        return PairingState.FAILED
                    return PairingState.SUCCESS

                elif res == 0:
                    return PairingState.SUCCESS
                elif res == 2:
                    return PairingState.FAILED
            except pexpect.TIMEOUT:
                return PairingState.TIMEOUT

    def trust(self, mac_address):
        try:
            self.send(f"trust {mac_address}", 4)
        except Exception as e:
            self.logger.error(e)
            return False
        else:
            res = self.process.expect(
                ["failed", "succeeded", pexpect.TIMEOUT],
                timeout=60,
            )
            return res == 1

    def remove(self, mac_address):
        """Remove paired device by mac address, return success of the operation."""
        try:
            self.send(f"remove {mac_address}", 3)
        except Exception as e:
            self.logger.error(e)
            return False
        else:
            res = self.process.expect(
                ["not available", "Device has been removed", pexpect.EOF]
            )
            return res == 1

    def connect(self, mac_address):
        """Try to connect to a device by mac address."""
        try:
            self.send(f"connect {mac_address}", 2)
        except Exception as e:
            self.logger.error(e)
            return False
        else:
            res = self.process.expect(
                ["Failed to connect", "Connection successful", pexpect.TIMEOUT],
                timeout=60,
            )
            return res == 1

    def disconnect(self, mac_address):
        """Try to disconnect to a device by mac address."""
        try:
            self.send(f"disconnect {mac_address}", 2)
        except Exception as e:
            self.logger.error(e)
            return False
        else:
            res = self.process.expect(
                ["Failed to disconnect", "Successful disconnected", pexpect.TIMEOUT],
                timeout=60,
            )
            return res == 1


# Enum class with the possible pairing states
class PairingState:
    SUCCESS = 0
    WRONG_PIN = 1
    TIMEOUT = 2
    FAILED = 3



import asyncio
import json
import logging
import random
import re
import string
import traceback
import functools
from typing import List, Optional

import requests

from fhempy.lib import utils
from .ezsp import EzspUtils
from .mini_miio import AsyncMiIO
from .shell import TelnetShell
from .xiaomi_cloud import MiCloud

DOMAIN = "xiaomi_gateway3"

_LOGGER = logging.getLogger(__name__)


# new miio adds colors to logs
RE_JSON1 = re.compile(b"msg:(.+) length:([0-9]+) bytes")
RE_JSON2 = re.compile(b"{.+}")


def extract_jsons(raw) -> List[bytes]:
    """There can be multiple concatenated json on one line. And sometimes the
    length does not match the message."""
    m = RE_JSON1.search(raw)
    if m:
        length = int(m[2])
        raw = m[1][:length]
    else:
        m = RE_JSON2.search(raw)
        raw = m[0]
    return raw.replace(b"}{", b"}\n{").split(b"\n")


def migrate_options(data):
    data = dict(data)
    options = {k: data.pop(k) for k in ("ble", "zha") if k in data}
    return {"data": data, "options": options}


async def check_mgl03(
    host: str, token: str, telnet_cmd: Optional[str]
) -> Optional[str]:
    # 1. try connect with telnet (custom firmware)?
    sh = TelnetShell()
    try:
        if await sh.connect(host):
            # 1.1. check token with telnet
            return None if await sh.get_token() == token else "wrong_token"

        if not telnet_cmd:
            return "cant_connect"

        # 2. try connect with miio
        miio = AsyncMiIO(host, token)
        info = await miio.info()

        # if info is None - devise doesn't answer on pings
        if info is None:
            return "cant_connect"

        # if empty info - device works but not answer on commands
        if not info:
            return "wrong_token"

        # 3. check if right model
        if info["model"] != "lumi.gateway.mgl03":
            return "wrong_model"

        raw = json.loads(telnet_cmd)
        # fw 1.4.6_0043+ won't answer on cmd without cloud, don't check answer
        await miio.send(raw["method"], raw.get("params"))

        # waiting for telnet to start
        await asyncio.sleep(1)

        if not await sh.connect(host):
            return "wrong_telnet"

        return None

    finally:
        await sh.close()


async def get_lan_key(host: str, token: str):
    device = AsyncMiIO(host, token)
    resp = await device.send("get_lumi_dpf_aes_key")
    if resp is None:
        return "Can't connect to gateway"
    if len(resp[0]) == 16:
        return resp[0]
    key = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(16)
    )
    resp = await device.send("set_lumi_dpf_aes_key", [key])
    if resp[0] == "ok":
        return key
    return "Can't update gateway key"


async def get_room_mapping(cloud: MiCloud, host: str, token: str):
    try:
        device = AsyncMiIO(host, token)
        local_rooms = await device.send("get_room_mapping")
        cloud_rooms = await cloud.get_rooms()
        result = ""
        for local_id, cloud_id in local_rooms:
            cloud_name = next(
                (p["name"] for p in cloud_rooms if p["id"] == cloud_id), "-"
            )
            result += f"\n- {local_id}: {cloud_name}"
        return result

    except:
        return "Can't get from cloud"


async def get_bindkey(cloud: MiCloud, did: str):
    bindkey = await cloud.get_bindkey(did)
    if bindkey is None:
        return "Can't get from cloud"
    if bindkey.endswith("FFFFFFFF"):
        return "Not needed"
    return bindkey


def reverse_mac(s: str):
    return f"{s[10:]}{s[8:10]}{s[6:8]}{s[4:6]}{s[2:4]}{s[:2]}"


NCP_URL = "https://master.dl.sourceforge.net/project/mgl03/zigbee/%s?viasf=1"


def flash_zigbee_firmware(
    host: str, ports: list, fw_url: str, fw_ver: str, fw_port=0, force=False
):
    """
    param host: gateway host
    param ports: one or multiple ports with different speeds, first port
        should be 115200
    param fw_url: url to firmware file
    param fw_ver: firmware version, checks before and after flash
    param fw_port: optional, port with firmware speed if it is not 115200
    param second_port: optional, second port if current firmware may have
        different speed
    param force: skip check firmware version before flash
    return: True if NCP firmware version equal to fw_ver
    """

    # we can flash NCP only in boot mode on speed 115200
    # but NCP can work on another speed, so we need to try both of them
    # work with 115200 on port 8115, and with 38400 on port 8038
    _LOGGER.debug(f"Try to update Zigbee NCP to version {fw_ver}")

    if isinstance(ports, int):
        ports = [ports]

    utils = EzspUtils()

    try:
        # try to find right speed from the list
        for port in ports:
            utils.connect(host, port)
            state = utils.state()
            if state:
                break
            utils.close()
        else:
            raise RuntimeError

        if state == "normal":
            if fw_ver in utils.version and not force:
                _LOGGER.info("No need to flash")
                return True
            _LOGGER.info(f"NCP state: {state}, version: {utils.version}")
            utils.launch_boot()
            state = utils.state()

        _LOGGER.info(f"NCP state: {state}, version: {utils.version}")

        # should be in boot
        if state != "boot":
            return False

        r = requests.get(fw_url)
        assert r.status_code == 200, r.status_code

        assert utils.flash_and_close(r.content)

        utils.connect(host, ports[0])
        utils.reboot_and_close()

        utils.connect(host, fw_port or ports[0])
        state = utils.state()
        _LOGGER.info(f"NCP state: {state}, version: {utils.version}")
        return fw_ver in utils.version

    except:
        _LOGGER.error(f"NCP flash error: {traceback.format_exc(1)}")
        return False

    finally:
        utils.close()


async def update_zigbee_firmware(host: str, custom: bool):
    """Update zigbee firmware for both ZHA and zigbee2mqtt modes"""

    sh = TelnetShell()
    try:
        if not await sh.connect(host) or not await sh.run_zigbee_flash():
            return False
    except:
        pass
    finally:
        await sh.close()

    await asyncio.sleep(0.5)

    args = (
        [
            host,
            [8115, 8038],
            NCP_URL % "mgl03_ncp_6_7_10_b38400_sw.gbl",
            "v6.7.10",
            8038,
        ]
        if custom
        else [
            host,
            [8115, 8038],
            NCP_URL % "ncp-uart-sw_mgl03_6_6_2_stock.gbl",
            "v6.6.2",
            8115,
        ]
    )

    for _ in range(3):
        if await utils.run_blocking(functools.partial(flash_zigbee_firmware, *args)):
            return True
    return False
